# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTIBILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

bl_info = {
    "name": "Combined Render",
    "author": "HDRdesign.Portugal",
    "version": (1, 0),
    "blender": (4, 0, 0),
    "location": "View3D > N-Panel",
    "description": "Renders 4 camera images and a combined image",
    "warning": "",
    "doc_url": "",
    "category": "Render",
}

##########################################
## IMPORTS
##########################################
import bpy
import os
import bpy.utils.previews
import subprocess
from bpy.props import StringProperty, BoolProperty, PointerProperty, IntProperty
from bpy.types import PropertyGroup, Operator, Panel, AddonPreferences
from mathutils import Vector

##############################################
## FUNCTIONS
##############################################
stored_scene_name = None
stored_main_scene_name = ""

def update_filename(scene, context):
    if 'WinMan' in bpy.context.window_manager:
        output_path = bpy.data.scenes[scene.name].render.filepath
        custom_name = bpy.context.window_manager['WinMan'].shared_custom_image_name

        if custom_name:
            output_path = output_path.replace(scene.name, custom_name)

        bpy.data.scenes[scene.name].render.filepath = output_path

bpy.types.WindowManager.shared_render_path = StringProperty(
    name="",
    subtype='DIR_PATH',
    default="//render_output",
    description="- Select the path to store the rendered images"
)
bpy.types.WindowManager.shared_custom_image_name = StringProperty(
    name="",
    default="Image",
    update=update_filename,
    description="- Add a name to the images as prefix"
)
bpy.types.WindowManager.shared_resolution_x = IntProperty(
    name="Shared Resolution X",
    default=1920,
    min=1,
)
bpy.types.WindowManager.shared_resolution_y = IntProperty(
    name="Shared Resolution Y",
    default=1080,
    min=1,
)

#########################################################################
# Function to link the new added nodes with the 'Set Render Output' button
#########################################################################
def link_nodes(context, scene):
    tree = scene.node_tree

    # Find the 'File Output' node
    for node in tree.nodes:
        if node.type == 'OUTPUT_FILE':
            file_output_node = node
            break

    # Find the 'Alpha Over.002' node
    alpha_over_002_node = None
    for node in tree.nodes:
        if node.name == 'Alpha Over.002':
            alpha_over_002_node = node
            break

    # Find the 'Render Layers' node
    render_layers_node = None
    for node in tree.nodes:
        if node.name == 'Render Layers':
            render_layers_node = node
            break

    # Find the 'Render Layers.001' node
    render_layers_001_node = None
    for node in tree.nodes:
        if node.name == 'Render Layers.001':
            render_layers_001_node = node
            break

    # Find the 'Render Layers.002' node
    render_layers_002_node = None
    for node in tree.nodes:
        if node.name == 'Render Layers.002':
            render_layers_002_node = node
            break 

    # Find the 'Render Layers.003' node
    render_layers_003_node = None
    for node in tree.nodes:
        if node.name == 'Render Layers.003':
            render_layers_003_node = node
            break   
        
    # Find the 'Scale' node
    scale_node = None
    for node in tree.nodes:
        if node.name == 'Scale':
            scale_node = node
            break 
        
    # Find the 'Scale.001' node
    scale_001_node = None
    for node in tree.nodes:
        if node.name == 'Scale.001':
            scale_001_node = node
            break
        
    # Find the 'Scale.002' node
    scale_002_node = None
    for node in tree.nodes:
        if node.name == 'Scale.002':
            scale_002_node = node
            break          

    # Find the 'Scale.003' node
    scale_003_node = None
    for node in tree.nodes:
        if node.name == 'Scale.003':
            scale_003_node = node
            break   
        
    # Find the 'Translate' node
    Translate_node = None
    for node in tree.nodes:
        if node.name == 'Translate':
            Translate_node = node
            break 
        
    # Find the 'Translate.001' node
    Translate_001_node = None
    for node in tree.nodes:
        if node.name == 'Translate.001':
            Translate_001_node = node
            break
        
    # Find the 'Translate.002' node
    Translate_002_node = None
    for node in tree.nodes:
        if node.name == 'Translate.002':
            Translate_002_node = node
            break

    # Find the 'Translate.003' node
    Translate_003_node = None
    for node in tree.nodes:
        if node.name == 'Translate.003':
            Translate_003_node = node
            break
    
    # Find the 'Alpha Over' node
    alpha_over_node = None
    for node in tree.nodes:
        if node.name == 'Alpha Over':
            alpha_over_node = node
            break          
    
    # Find the 'Alpha Over.001' node
    alpha_over_001_node = None
    for node in tree.nodes:
        if node.name == 'Alpha Over.001':
            alpha_over_001_node = node
            break      
  
    # Create a connection between the 'File Output' node and the 'Alpha Over.002' node
    if file_output_node and alpha_over_002_node:
        tree.links.new(file_output_node.inputs[0], alpha_over_002_node.outputs[0])
        tree.links.new(render_layers_node.outputs["Image"], file_output_node.inputs[1])
        tree.links.new(render_layers_001_node.outputs["Image"], file_output_node.inputs[2])
        tree.links.new(render_layers_002_node.outputs["Image"], file_output_node.inputs[3])
        tree.links.new(render_layers_003_node.outputs["Image"], file_output_node.inputs[4])

        print("Link between 'File Output' and 'Alpha Over' node created successfully.")
    else:
        print("Failed to create link between 'File Output' and 'Alpha Over' node.")

    # Create a connection between the 'scale' nodes and the 'Translate' nodes
    if scale_node and Translate_node:
        tree.links.new(Translate_node.inputs[0], scale_node.outputs["Image"])
        tree.links.new(Translate_001_node.inputs[0], scale_001_node.outputs["Image"])
        tree.links.new(Translate_002_node.inputs[0], scale_002_node.outputs["Image"])
        tree.links.new(Translate_003_node.inputs[0], scale_003_node.outputs["Image"])
        
    # Create a connection between the 'Translate_node' nodes and the 'Alpha Over' nodes
    if Translate_node and alpha_over_node:
        tree.links.new(alpha_over_node.inputs[1], Translate_node.outputs["Image"])
        tree.links.new(alpha_over_node.inputs[2], Translate_001_node.outputs["Image"])
        tree.links.new(alpha_over_001_node.inputs[1], Translate_002_node.outputs["Image"])
        tree.links.new(alpha_over_001_node.inputs[2], Translate_003_node.outputs["Image"])
    
    # Create a connection between the 'Alpha Over' nodes    
        tree.links.new(alpha_over_002_node.inputs[1], alpha_over_node.outputs["Image"])
        tree.links.new(alpha_over_002_node.inputs[2], alpha_over_001_node.outputs["Image"])


###########
# Function to link the 'File Output' node and the 'Alpha Over.002' node
###########
def link_ftlr_nodes(context):
    tree = bpy.context.scene.node_tree

    # Find the 'File Output' node 
    for node in tree.nodes:
        if node.type == 'OUTPUT_FILE':
            file_output_node = node
            break

    # Find the 'Alpha Over.002' node
    alpha_over_002_node = None
    for node in tree.nodes:
        if node.name == 'Alpha Over.002':
            alpha_over_002_node = node
            break

    # Find the 'Render Layers' node
    render_layers_node = None
    for node in tree.nodes:
        if node.name == 'Render Layers':
            render_layers_node = node
            break

    # Find the 'Render Layers.001' node
    render_layers_001_node = None
    for node in tree.nodes:
        if node.name == 'Render Layers.001':
            render_layers_001_node = node
            break

    # Find the 'Render Layers.002' node
    render_layers_002_node = None
    for node in tree.nodes:
        if node.name == 'Render Layers.002':
            render_layers_002_node = node
            break 

    # Find the 'Render Layers.003' node
    render_layers_003_node = None
    for node in tree.nodes:
        if node.name == 'Render Layers.003':
            render_layers_003_node = node
            break  
        
    # Find the 'Scale' node
    scale_node = None
    for node in tree.nodes:
        if node.name == 'Scale':
            scale_node = node
            break 
        
    # Find the 'Scale.001' node
    scale_001_node = None
    for node in tree.nodes:
        if node.name == 'Scale.001':
            scale_001_node = node
            break
        
    # Find the 'Scale.002' node
    scale_002_node = None
    for node in tree.nodes:
        if node.name == 'Scale.002':
            scale_002_node = node
            break          

    # Find the 'Scale.003' node
    scale_003_node = None
    for node in tree.nodes:
        if node.name == 'Scale.003':
            scale_003_node = node
            break   
    
    # Find the 'Translate' node
    Translate_node = None
    for node in tree.nodes:
        if node.name == 'Translate':
            Translate_node = node
            break 
        
    # Find the 'Translate.001' node
    Translate_001_node = None
    for node in tree.nodes:
        if node.name == 'Translate.001':
            Translate_001_node = node
            break
        
    # Find the 'Translate.002' node
    Translate_002_node = None
    for node in tree.nodes:
        if node.name == 'Translate.002':
            Translate_002_node = node
            break

    # Find the 'Translate.003' node
    Translate_003_node = None
    for node in tree.nodes:
        if node.name == 'Translate.003':
            Translate_003_node = node
            break
    
    # Find the 'Alpha Over' node
    alpha_over_node = None
    for node in tree.nodes:
        if node.name == 'Alpha Over':
            alpha_over_node = node
            break          
    
    # Find the 'Alpha Over.001' node
    alpha_over_001_node = None
    for node in tree.nodes:
        if node.name == 'Alpha Over.001':
            alpha_over_001_node = node
            break 
        
    # Create a connection between the 'File Output' node and the 'Alpha Over.002' node
    if file_output_node and alpha_over_002_node:
        tree.links.new(file_output_node.inputs[0], alpha_over_002_node.outputs[0])
        tree.links.new(render_layers_node.outputs["Image"], file_output_node.inputs[1])
        tree.links.new(render_layers_001_node.outputs["Image"], file_output_node.inputs[2])
        tree.links.new(render_layers_002_node.outputs["Image"], file_output_node.inputs[3])
        tree.links.new(render_layers_003_node.outputs["Image"], file_output_node.inputs[4])

        print("Link between 'File Output' and 'Alpha Over' node created successfully.")
    else:
        print("Failed to create link between 'File Output' and 'Alpha Over' node.")

    # Create a connection between the 'scale' nodes and the 'Translate' nodes
    if scale_node and Translate_node:
        tree.links.new(Translate_node.inputs[0], scale_node.outputs["Image"])
        tree.links.new(Translate_001_node.inputs[0], scale_001_node.outputs["Image"])
        tree.links.new(Translate_002_node.inputs[0], scale_002_node.outputs["Image"])
        tree.links.new(Translate_003_node.inputs[0], scale_003_node.outputs["Image"])
        
    # Create a connection between the 'Translate_node' nodes and the 'Alpha Over' nodes
    if Translate_node and alpha_over_node:
        tree.links.new(alpha_over_node.inputs[1], Translate_node.outputs["Image"])
        tree.links.new(alpha_over_node.inputs[2], Translate_001_node.outputs["Image"])
        tree.links.new(alpha_over_001_node.inputs[1], Translate_002_node.outputs["Image"])
        tree.links.new(alpha_over_001_node.inputs[2], Translate_003_node.outputs["Image"])
         
    # Create a connection between the 'Alpha Over' nodes    
        tree.links.new(alpha_over_002_node.inputs[1], alpha_over_node.outputs["Image"])
        tree.links.new(alpha_over_002_node.inputs[2], alpha_over_001_node.outputs["Image"])


bpy.types.WindowManager.shared_ftlr_resolution_x = IntProperty(
    name="Shared Resolution X",
    default=1920,
    min=1,
)
bpy.types.WindowManager.shared_ftlr_resolution_y = IntProperty(
    name="Shared Resolution Y",
    default=1080,
    min=1,
)

def update_ftlr_filename(scene, context):
    if 'WinMan' in bpy.context.window_manager:
        output_path = bpy.data.scenes[scene.name].render.filepath
        custom_name = bpy.context.window_manager['WinMan'].shared_ftlr_custom_image_name
        if custom_name:
            output_path = output_path.replace(scene.name, custom_name)
        bpy.data.scenes[scene.name].render.filepath = output_path     

bpy.types.WindowManager.shared_ftlr_render_path = StringProperty(
    name="",
    subtype='DIR_PATH',
    default="//render_output",
    description="- Select the path to store the rendered images"
)
bpy.types.WindowManager.shared_ftlr_custom_image_name = StringProperty(
    name="",
    default="Image",
    update=update_ftlr_filename,
    description="- Add a name to the images as prefix\n- Default = Image"
)


#######################
## PANEL 1
#######################                     
class CORE_PT_cameras(Panel):
    bl_label = "4 SCENES CAMERA"
    bl_idname = "CORE_PT_cameras"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'ComRen'
    bl_order = 0
    
    def draw_header(self, context):
        layout = self.layout   
        layout.template_icon(icon_value=bpy.context.scene.Combined_Render_icons['FOURSCENES'].icon_id, scale=1.0)

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        current_scene = context.scene
     
        cameras_present = any(obj.get("added_camera", False) for obj in bpy.data.objects if obj.type == 'CAMERA')

        if cameras_present:
            layout.label(text="** CAMERAS ADDED **")
        else:
            row = layout.row()
            row.operator("scene.multi_copy_scene", text="Add 3 Scene Copies", icon_value=bpy.context.scene.Combined_Render_icons['3COPY'].icon_id) 
            row.scale_y = 1.3   
                
        box = layout.box()
        row = box.row() 
        row.label(text="Selected 4 Scenes:")
        added_cams = 0
        for scene in bpy.data.scenes:
            if scene.name in ['FrontCamera', 'TopCamera', 'LeftCamera', 'RightCamera']:
                continue
            
            if scene.use_for_cameras and added_cams < 4:
                scene_row = box.row(align=True)
                scene_row.prop(scene, "use_for_cameras", text=scene.name)

                sub_box = scene_row.box()
                sub_box.operator("scene.switch_to_scene", text="To Scene", depress=scene == context.scene).scene_name = scene.name
                sub_box.scale_x = 1.0 

                linked_camera_selected = any(
                    obj.type == 'CAMERA' and obj.select_get() and not obj.hide_get() and obj.get("added_camera", False) for obj in scene.collection.all_objects
                )
                if linked_camera_selected:
                    camera = context.scene.camera
                    if camera and camera.type == 'CAMERA':
                        scene_row = box.row()
                        scene_row.prop(camera.data.dof, "use_dof", text="Depth of Field", toggle=True, emboss=False,
                                       icon='RADIOBUT_ON' if camera.data.dof.use_dof else 'RADIOBUT_OFF')
                        scene_row = box.row()    
                        # Check if Depth of Field is enabled
                        if camera.data.dof.use_dof:
                            scene_row.label(text="Focus Object:")
                            scene_row = box.row()
                            scene_row.prop_search(camera.data.dof, "focus_object", scene, "objects", text="")

                        scene_row = box.row()
                        scene_row.prop(camera.data, "show_limits", text="Show Limits", toggle=True, emboss=False,
                                    icon='RADIOBUT_ON' if camera.data.show_limits else 'RADIOBUT_OFF')
                added_cams += 1
            else:
                box.prop(scene, "use_for_cameras", text=scene.name, toggle=False, icon='LOCKED')
 
        # Check if there are 4 scenes in the blend file
        if len(bpy.data.scenes) == 4:
            row = layout.row() 
            row.label(text= "Camera Distance (mm):")
            layout.prop(current_scene, "cam_empty_distance", text="") 

        row = layout.row()   
        row.label(text= "Add Cameras To Scenes:") 
        row = layout.row()    
        row.operator("object.add_cameras", text="Add Cameras", icon_value=bpy.context.scene.Combined_Render_icons['ADDCAM'].icon_id)
        row.scale_y = 2.0    
        row = layout.row()  
        
        if any(obj.get("added_camera", False) for obj in bpy.data.objects if obj.type == 'CAMERA'):
            
            row.operator("object.delete_cameras", text="Delete Cameras",icon_value=bpy.context.scene.Combined_Render_icons['DELCAM'].icon_id)
            row.scale_y = 1.5     
            row = layout.row()      

            settings_box = layout.box()
            # Image Name
            row = settings_box.row()
            row.label(text= "Image Name:")
            # Display a text field for the custom image name
            settings_box.prop(context.window_manager, "shared_custom_image_name", text="")

            # Render Output Path
            row = settings_box.row()
            row.label(text= "Render Output Path:")
            # Adding a path selector for the shared render path
            settings_box.prop(context.window_manager, "shared_render_path", text="")

            # Resolution
            row = settings_box.row()
            row.label(text= "Resolution:", icon_value=bpy.context.scene.Combined_Render_icons['RESOL'].icon_id)
            settings_box.prop(context.window_manager, "shared_resolution_x", text="Resolution X")
            settings_box.prop(context.window_manager, "shared_resolution_y", text="Resolution Y")
            
            # Toggle resolution Default/Insta
            row = settings_box.row()
            row.operator("render.set_insta_resolution", text="Insta / Default")

            # Apply Settings
            row = settings_box.row()
            row.label(text= "Apply Settings:")
            row = settings_box.row()
            # Add button to set render output for the current project and change the image name
            row.operator("render.custom_render_output", text="Output Settings", icon_value=bpy.context.scene.Combined_Render_icons['SETTINGS'].icon_id) 
            row.scale_y = 1.5     

            # Open Folder After Render Checkbox
            row = settings_box.row()
            row.prop(context.window_manager, "open_folder_after_render", text="Open Folder After Render")

            # Render Images
            row = settings_box.row()
            row.label(text= "Render Images:")
            row = settings_box.row()
            # Button to render the 4 views
            row.operator("render.render_button", icon="RENDER_STILL")
            row.scale_y = 2.3     
            row = settings_box.row()

    
#######################
## PANEL 2
####################### 
class CORE_PT_ftlr_cameras(Panel):
    bl_label = "OBJECT CAMERAS"
    bl_idname = "CORE_PT_ftlr_cameras"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'ComRen'
    bl_order = 1

    def draw_header(self, context):
        layout = self.layout   
        layout.template_icon(icon_value=bpy.context.scene.Combined_Render_icons['FTLR'].icon_id, scale=1.0)

    def draw(self, context):
        layout = self.layout
        scene = context.scene

        row = layout.row()
        row.label(text="Add View Cameras:")

        row = layout.row()
        # Button to add cameras
        row.operator("object.add_four_cameras", text="Add View Cameras", icon_value=bpy.context.scene.Combined_Render_icons['ADDCAM'].icon_id)
        row.scale_y = 2.3  

        # Get scenes with specific names
        scene_names = ['FrontCamera', 'TopCamera', 'LeftCamera', 'RightCamera']
        scenes = [scene for scene in bpy.data.scenes if scene.name in scene_names]

        if all(scene_name in bpy.data.scenes for scene_name in ["FrontCamera", "TopCamera", "RightCamera", "LeftCamera"]):
            box = layout.box()
            row = box.row()
            row.label(text="Camera Views:")

        # Display scene names and buttons
        for scene in scenes:
            if scene.name in scene_names:
                row = box.row(align=True)
                is_active = context.scene == scene

                if scene.name == 'FrontCamera':
                    row.operator("scene.switch_to_front_camera", text="To Front Camera", icon='RADIOBUT_ON' if is_active else 'RADIOBUT_OFF', depress=scene == context.scene).scene_name = scene.name    
                elif scene.name == 'TopCamera':
                    row.operator("scene.switch_to_top_camera", text="To Top Camera", icon='RADIOBUT_ON' if is_active else 'RADIOBUT_OFF', depress=scene == context.scene).scene_name = scene.name   
                elif scene.name == 'LeftCamera':
                    row.operator("scene.switch_to_left_camera", text="To Left Camera", icon='RADIOBUT_ON' if is_active else 'RADIOBUT_OFF', depress=scene == context.scene).scene_name = scene.name   
                elif scene.name == 'RightCamera':
                    row.operator("scene.switch_to_right_camera", text="To Right Camera", icon='RADIOBUT_ON' if is_active else 'RADIOBUT_OFF', depress=scene == context.scene).scene_name = scene.name   
                    
                # Display options for the selected camera
                if is_active:
                    camera_name = scene.name
                    if camera_name in scene.objects and scene.objects[camera_name].type == 'CAMERA':
                        camera_object = scene.objects[camera_name]

                        row = box.row(align=True)
                        row.prop(camera_object.data.dof, "use_dof", text="Depth of Field", toggle=True, emboss=False,
                                 icon='RADIOBUT_ON' if camera_object.data.dof.use_dof else 'RADIOBUT_OFF')

                        if camera_object.data.dof.use_dof:
                            row = box.row(align=True)
                            row.label(text="Focus Object:")
                            row = box.row()
                            row.prop_search(camera_object.data.dof, "focus_object", bpy.context.scene, "objects",
                                            text="")

                        row = box.row(align=True)
                        row.prop(camera_object.data, "show_limits", text="Show Limits", toggle=True, emboss=False,
                                 icon='RADIOBUT_ON' if camera_object.data.show_limits else 'RADIOBUT_OFF')
                                       
        # Show the "Switch to SOURCE Scene" button only if the four scenes are added
        if all(scene_name in bpy.data.scenes for scene_name in ["FrontCamera", "TopCamera", "RightCamera", "LeftCamera"]):
            box = layout.box()
            row = box.row()
            row.operator("object.switch_to_main_scene", text="Switch To Source Scene")
            row.scale_y = 1.5

            row = box.row()
            row.operator("object.track_selected_object", text="Track Selected Object")

            row = box.row()
            row.operator("object.move_cursor_to_origin", text="Center Cameras")

            row = layout.row() 
            row = layout.row()
            # Button to delete cameras
            row.operator("object.delete_ftlr_cameras", text="Delete View Cameras",icon_value=bpy.context.scene.Combined_Render_icons['DELCAM'].icon_id)
            row.scale_y = 1.5 
            row = layout.row()
        
        # Show the Render settings button only if the four scenes are added
        if all(scene_name in bpy.data.scenes for scene_name in ["FrontCamera", "TopCamera", "RightCamera", "LeftCamera"]):
            added_scenes_box = layout.box()
            
            # Image Name
            row = added_scenes_box.row()
            row.label(text= "Image Name:") 
            # Display a text field for the custom image name
            added_scenes_box.prop(context.window_manager, "shared_ftlr_custom_image_name", text="")

            # Render Output Path
            row = added_scenes_box.row()
            row.label(text= "Render Output Path:") 
            # Adding a path selector for the render path
            added_scenes_box.prop(context.window_manager, "shared_ftlr_render_path", text="")

            # Resolution
            row = added_scenes_box.row()
            row.label(text= "Resolution:", icon_value=bpy.context.scene.Combined_Render_icons['RESOL'].icon_id)
            added_scenes_box.prop(context.window_manager, "shared_ftlr_resolution_x", text="Resolution X")
            added_scenes_box.prop(context.window_manager, "shared_ftlr_resolution_y", text="Resolution Y")

            # Toggle resolution Default/Insta
            row = added_scenes_box.row()
            row.operator("render.set_ftlr_insta_resolution", text="Insta / Default")

            # Apply Settings
            row = added_scenes_box.row()
            row.label(text= "Apply Settings:", icon_value=bpy.context.scene.Combined_Render_icons['SETTINGS'].icon_id)
            row = added_scenes_box.row()
            # Add button to set render output for the current project and change the image name
            row.operator("render.custom_ftlr_render_output", text="Output Settings")
            row.scale_y = 1.5
            
            # Open Folder After Render Checkbox
            row = added_scenes_box.row()
            row.prop(context.window_manager, "ftlr_open_folder_after_render", text="Open Folder After Render")

            # Render Images
            row = added_scenes_box.row() 
            row.label(text= "Render Images:") 
            row = added_scenes_box.row()
            # Button to render the 4 views
            row.operator("render.render_ftlr_button", icon="RENDER_STILL")
            row.scale_y = 2.3
            row = added_scenes_box.row()
        
##############################################
## OPERATORS
##############################################
   
class CORE_OT_SwitchScene(Operator):
    bl_idname = "scene.switch_to_scene"
    bl_label = "Switch to Scene"
    bl_description = "- Switches to the checked scene and selects the camera if added"
    
    scene_name : StringProperty()

    def execute(self, context):
        bpy.context.window.scene = bpy.data.scenes.get(self.scene_name)

        scene = bpy.context.scene
        for obj in scene.collection.all_objects:
            if obj.type == 'CAMERA' and obj.get("added_camera", False):
                obj.select_set(True)
                bpy.context.view_layer.objects.active = obj
                bpy.context.object.show_name = True
            else:
                obj.select_set(False)

        return {'FINISHED'}

class CORE_OT_MultiCopyScene(Operator):
    bl_idname = "scene.multi_copy_scene"
    bl_label = "Add 3 Scene Copies"
    bl_options = {'REGISTER', 'UNDO'}
    bl_description = "- Makes 3 copies of the active scene\n- If the same World properties are needed"

    def execute(self, context):
        global stored_scene_name
        active_scene = context.scene
        stored_scene_name = active_scene.name
        
        active_scene = bpy.context.scene
        if not active_scene:
            self.report({'ERROR'}, "No active scene found.")
            return {'CANCELLED'}
       
        active_scene.use_nodes = True

        # Make 3 full copies of the selected active scene
        for i in range(3):
            bpy.ops.scene.new(type='FULL_COPY')
            
        context.window.scene = bpy.data.scenes[stored_scene_name]

        return {'FINISHED'}
    
class CORE_OT_AddCameras(Operator):
    bl_idname = "object.add_cameras"
    bl_label = "Add Cameras"
    bl_options = {'REGISTER', 'UNDO'}
    bl_description = "- Adds a camera to the checked scenes.\n- Selecting 4 scenes will enable this button"
    
    @classmethod
    def poll(cls, context):
        non_specific_scenes = [scene for scene in bpy.data.scenes if scene.name not in ['FrontCamera', 'TopCamera', 'RightCamera', 'LeftCamera']]
        non_specific_scene_count = len(non_specific_scenes)

        return (
            not any(obj.get("added_camera", False) for obj in bpy.data.objects if obj.type == 'CAMERA')
            and non_specific_scene_count >= 4
        )
        
    def setup_compositor(self, scene, added_cams_scenes):
        scene.use_nodes = True
        tree = scene.node_tree
        # Remove existing nodes
        for node in tree.nodes:
            tree.nodes.remove(node)
        scene.node_tree.links.clear()
        scene.node_tree.nodes.clear()

        render_layer_nodes = []
        scale_nodes = []
        translate_nodes = []

        for i, scene_name in enumerate(added_cams_scenes):
            render_layer_node = scene.node_tree.nodes.new(type="CompositorNodeRLayers")
            render_layer_node.location = (-200, -i * 200)

            render_layer_node.scene = bpy.data.scenes[scene_name]
            render_layer_nodes.append(render_layer_node)

            scale_node = scene.node_tree.nodes.new(type="CompositorNodeScale")
            scale_node.location = (200, -i * 200)
            scale_node.inputs[1].default_value = 0.5
            scale_node.inputs[2].default_value = 0.5
            scale_nodes.append(scale_node)

            scene.node_tree.links.new(render_layer_node.outputs["Image"], scale_node.inputs[0])

        scale_values = [(0.5, 0.5, 0), (0.5, 0.5, 0), (0.5, 0.5, 0), (0.5, 0.5, 0)]

        for i, scale_node in enumerate(scale_nodes):
            scale_node.inputs[1].default_value = scale_values[i][0]  
            scale_node.inputs[2].default_value = scale_values[i][1]  

            translate_node = scene.node_tree.nodes.new(type="CompositorNodeTranslate")
            translate_node.location = (400, -i * 200)
            translate_nodes.append(translate_node)

            scene.node_tree.links.new(scale_node.outputs["Image"], translate_node.inputs[0])

        shared_resolution_x = bpy.context.window_manager.shared_resolution_x
        shared_resolution_y = bpy.context.window_manager.shared_resolution_y

        translate_values = [
            (-shared_resolution_x / 4, shared_resolution_y / 4, 0),
            (shared_resolution_x / 4, shared_resolution_y / 4, 0),
            (-shared_resolution_x / 4, -shared_resolution_y / 4, 0),
            (shared_resolution_x / 4, -shared_resolution_y / 4, 0),
        ]

        for i, translate_node in enumerate(translate_nodes):
            translate_node.inputs[1].default_value = translate_values[i][0]
            translate_node.inputs[2].default_value = translate_values[i][1]

        alpha_over_node1 = scene.node_tree.nodes.new(type="CompositorNodeAlphaOver")
        alpha_over_node1.location = (600, -len(added_cams_scenes) * 100)

        scene.node_tree.links.new(translate_nodes[0].outputs["Image"], alpha_over_node1.inputs[1])
        scene.node_tree.links.new(translate_nodes[1].outputs["Image"], alpha_over_node1.inputs[2])

        alpha_over_node2 = scene.node_tree.nodes.new(type="CompositorNodeAlphaOver")
        alpha_over_node2.location = (800, -len(added_cams_scenes) * 100)

        scene.node_tree.links.new(translate_nodes[2].outputs["Image"], alpha_over_node2.inputs[1])
        scene.node_tree.links.new(translate_nodes[3].outputs["Image"], alpha_over_node2.inputs[2])

        alpha_over_node3 = scene.node_tree.nodes.new(type="CompositorNodeAlphaOver")
        alpha_over_node3.location = (1000, -len(added_cams_scenes) * 100)

        scene.node_tree.links.new(alpha_over_node1.outputs["Image"], alpha_over_node3.inputs[1])
        scene.node_tree.links.new(alpha_over_node2.outputs["Image"], alpha_over_node3.inputs[2])

        viewer_node = scene.node_tree.nodes.new(type="CompositorNodeViewer")
        viewer_node.location = (1200, -len(added_cams_scenes) * 100)

        scene.node_tree.links.new(alpha_over_node3.outputs["Image"], viewer_node.inputs[0])

        composite_node = scene.node_tree.nodes.new(type="CompositorNodeComposite")
        composite_node.location = (1400, -len(added_cams_scenes) * 100)

        scene.node_tree.links.new(alpha_over_node3.outputs["Image"], composite_node.inputs[0])

        file_output_node = scene.node_tree.nodes.new(type="CompositorNodeOutputFile")
        file_output_node.location = (1600, -len(added_cams_scenes) * 100)
        file_output_node.label = 'InitialOutput'

        scene.node_tree.links.new(alpha_over_node3.outputs["Image"], file_output_node.inputs[0])

        file_slots = file_output_node.file_slots
        for _ in range(4):
            file_slots.new("RenderLayer")

        for i, render_layer_node in enumerate(render_layer_nodes):
            output_index = i + 1
            scene.node_tree.links.new(render_layer_node.outputs["Image"], file_output_node.inputs[output_index])

        print(f"Compositor node setup created in {scene.name}.")

        return translate_nodes

    def execute(self, context):
        global stored_scene_name
        active_scene = context.scene
        stored_scene_name = active_scene.name

        cursor_location = bpy.context.scene.cursor.location.copy()
        distance_mm = int(context.scene.cam_empty_distance)
        distance = distance_mm / 1000.0
        added_cameras = 0
        added_cams_scenes = []

        for scene in bpy.data.scenes:
            if scene.use_nodes is False:
                scene.use_nodes = True
                self.report({'INFO'}, f"Use Nodes enabled in {scene.name}")
            else:
                self.report({'WARNING'}, f"Use Nodes is already enabled in {scene.name}")
            
            if scene.name in ['FrontCamera', 'TopCamera', 'LeftCamera', 'RightCamera']:
                continue
            
            if scene.use_for_cameras and added_cameras < 4:
                cam_collection_name = "4CamViews_" + scene.name
                cam_collection = bpy.data.collections.get(cam_collection_name)
                if not cam_collection:
                    cam_collection = bpy.data.collections.new(cam_collection_name)
                    scene.collection.children.link(cam_collection)

                empty = bpy.data.objects.new("CamEmpty_" + scene.name, None)
                empty.location = cursor_location
                cam_collection.objects.link(empty)

                camera_data = bpy.data.cameras.new("Cam_" + scene.name)
                camera = bpy.data.objects.new("Cam_" + scene.name, camera_data)
                cam_collection.objects.link(camera)
                camera.location = cursor_location + Vector((distance, 0, 0))
                scene.camera = camera
                camera["added_camera"] = True
                track_constraint = camera.constraints.new(type='TRACK_TO')
                track_constraint.target = empty
                track_constraint.track_axis = 'TRACK_NEGATIVE_Z'
                track_constraint.up_axis = 'UP_Y'

                added_cams_scenes.append(scene.name)
                added_cameras += 1

        bpy.context.scene["added_cams_scenes"] = added_cams_scenes
        print("Added cameras to scenes:", added_cams_scenes)

        scene = bpy.data.scenes.get(stored_scene_name)

        if scene:
            translate_nodes = self.setup_compositor(scene, added_cams_scenes)
        else:
            translate_nodes = []

        context.scene.cameras_added = True
        
      # Get the active scene
        active_scene = context.scene
      # Store the active scene name
        stored_scene_name = active_scene.name
       
        self.report({"INFO"}, "Camera added to selected scenes")
        
        return {'FINISHED'}

class CORE_OT_SetInstaResolution(Operator):
    bl_idname = "render.set_insta_resolution"
    bl_label = "Toggle Resolution"
    bl_description = "- Toggle resolution between 1080x1350 and 1920x1080\n- 1920x1080 is default"
    
    def execute(self, context):
        # Toggle between resolutions
        if bpy.context.window_manager.shared_resolution_x == 1080 and bpy.context.window_manager.shared_resolution_y == 1350:
            bpy.context.window_manager.shared_resolution_x = 1920
            bpy.context.window_manager.shared_resolution_y = 1080
        else:
            bpy.context.window_manager.shared_resolution_x = 1080
            bpy.context.window_manager.shared_resolution_y = 1350
        return {'FINISHED'}

class CORE_RenderPaths(PropertyGroup):
    render_output_directory : bpy.props.StringProperty(
        name="Render Output Directory",
        subtype='DIR_PATH',
        default="//render_output",
    )

#########################################
# Operator to change the File Output path
######################################### 
    
def find_scene_with_output_node_label():
    for scene in bpy.data.scenes:
        if scene.use_nodes:
            for node in scene.node_tree.nodes:
                if node.type == 'OUTPUT_FILE' and (node.label == 'InitialOutput' or node.label == 'NewOutput'):
                    return scene
    return None

class CORE_OT_CustomRenderOutput(Operator):
    bl_idname = "render.custom_render_output"
    bl_label = "Set Render Output"
    bl_description = "- Confirm the image name, render output path and resolution "
    
    @classmethod
    def poll(cls, context):
        return any(obj.get("added_camera", False) for obj in bpy.data.objects if obj.type == 'CAMERA')

    def execute(self, context):
        active_scene = find_scene_with_output_node_label()
        if active_scene:
            
            global stored_scene_name
            stored_scene_name = active_scene.name
            bpy.context.window.scene = active_scene
        else:
            self.report({'ERROR'}, "No scene found with 'InitialOutput' or 'NewOutput' node.")
            return {'CANCELLED'}

        stored_scene = bpy.data.scenes.get(stored_scene_name)
        
        if stored_scene:
            stored_scene.render.filepath = bpy.context.window_manager.shared_render_path

        # Set the resolution 
        shared_resolution_x = bpy.context.window_manager.shared_resolution_x
        shared_resolution_y = bpy.context.window_manager.shared_resolution_y

        for scene_name in bpy.context.scene["added_cams_scenes"]:
            scene = bpy.data.scenes.get(scene_name)
            if scene:
                scene.render.resolution_x = shared_resolution_x
                scene.render.resolution_y = shared_resolution_y
                scene.render.resolution_percentage = 100

        bpy.context.scene.use_nodes = True
        tree = stored_scene.node_tree
        tree_type = 'COMPOSITING'
        
        for node in tree.nodes:
            if node.bl_idname == 'CompositorNodeTranslate':
                tree.nodes.remove(node)

        for node in tree.nodes:
            if node.type == 'OUTPUT_FILE' and (node.label == 'InitialOutput' or node.label == 'NewOutput'): 
                tree.nodes.remove(node)

        # Add a new 'File Output' node to the stored scene
        new_output_file_node = tree.nodes.new(type='CompositorNodeOutputFile')
        new_output_file_node.label = 'NewOutput'
        new_output_file_node.location = (1200, -600)
        new_output_file_node.base_path = bpy.context.window_manager.shared_render_path
        new_output_file_node.file_slots[0].path = bpy.context.window_manager.shared_custom_image_name + "_Combined_Image_"

        # Add four new inputs to the File Output node with specific names
        input_names = ["Cam01_", "Cam02_", "Cam03_", "Cam04_"]
        for input_name in input_names:
            new_output_file_node.file_slots.new(bpy.context.window_manager.shared_custom_image_name + "_" + input_name)

        # Add four 'Translate' nodes
        translate_nodes = []
        translate_values = [
            (-shared_resolution_x / 4, shared_resolution_y / 4, 0),
            (shared_resolution_x / 4, shared_resolution_y / 4, 0),
            (-shared_resolution_x / 4, -shared_resolution_y / 4, 0),
            (shared_resolution_x / 4, -shared_resolution_y / 4, 0),
        ]

        for i in range(4):
            translate_node = tree.nodes.new(type='CompositorNodeTranslate')
            translate_node.location = (400, -i * 200)
            translate_node.inputs[1].default_value = translate_values[i][0]  # X translation
            translate_node.inputs[2].default_value = translate_values[i][1]  # Y translation
            translate_nodes.append(translate_node)

        # Link nodes
        link_nodes(context, stored_scene)

        # Set the active scene to the stored scene
        context.window.scene = stored_scene

        self.report({"INFO"}, "Output settings adjusted")

        return {'FINISHED'}

class CORE_OT_RenderButton(bpy.types.Operator):
    bl_idname = "render.render_button"
    bl_label = "RENDER"
    bl_description = "Renders 4 images and a combined image to the selected path"

    @classmethod
    def poll(cls, context):
        return any(node.type == 'OUTPUT_FILE' and node.label == 'NewOutput' for node in bpy.context.scene.node_tree.nodes)

    def execute(self, context):
        global stored_scene_name
        if stored_scene_name in bpy.data.scenes:
            context.window.scene = bpy.data.scenes[stored_scene_name]
        
        # Register render complete callback function
        bpy.app.handlers.render_complete.append(self.open_render_output_directory)

        bpy.ops.render.render("INVOKE_DEFAULT")

        return {'RUNNING_MODAL'}

    def open_render_output_directory(self, *args):
        # This function is called when rendering is completed
        print("Rendering completed. Opening render output directory...")
        render_output_directory = bpy.path.abspath(bpy.context.window_manager.shared_render_path)

        try:
            if bpy.context.window_manager.open_folder_after_render:
                if os.name == 'nt':  # Windows
                    subprocess.Popen(['explorer', render_output_directory], shell=True)
                elif os.name == 'posix':  # Linux and macOS
                    subprocess.Popen(['xdg-open', render_output_directory])
                else:
                    print("Unsupported operating system.")
        except Exception as e:
            print(f"Error opening file explorer: {e}")

        # Unregister the callback function
        bpy.app.handlers.render_complete.remove(self.open_render_output_directory)
        
        return {'FINISHED'}
    
    def bl_rna_update(self, context):
        bpy.app.handlers.render_complete.append(self.open_render_output_directory) 

          
class CORE_OT_DeleteCameras(Operator):
    bl_idname = "object.delete_cameras"
    bl_label = "Delete Cameras"
    bl_options = {'REGISTER', 'UNDO'}
    bl_description = "- Deletes the added cameras of the selected scenes"

    @classmethod
    def poll(cls, context):
        return any(obj.get("added_camera", False) for obj in bpy.data.objects if obj.type == 'CAMERA')

    def execute(self, context):
        global stored_scene_name
        if stored_scene_name in bpy.data.scenes:
            context.window.scene = bpy.data.scenes[stored_scene_name]
            
        bpy.context.scene.camera = None

        objects_to_remove = [obj for obj in bpy.data.objects if (obj.type == 'CAMERA' and obj.get("added_camera", False)) or (obj.type == 'EMPTY' and obj.name.startswith("CamEmpty"))]
        bpy.data.batch_remove(objects_to_remove)

        collections_to_remove = [collection for collection in bpy.data.collections if collection.name.startswith("4CamViews_")]
        bpy.data.batch_remove(collections_to_remove)

        stored_scene = bpy.data.scenes.get(stored_scene_name)
        if stored_scene:
            print(f"Found master scene: {stored_scene_name}")
            self.delete_compositor_node_tree(stored_scene) 
        else:
            print(f"Master scene not found: {stored_scene_name}")     
                  
        bpy.ops.outliner.orphans_purge(do_local_ids=True, do_linked_ids=True, do_recursive=True)

        return {'FINISHED'}
    
    def delete_compositor_node_tree(self, scene):
        
        node_tree = scene.node_tree
        if node_tree:
            for node in node_tree.nodes:
                node_tree.nodes.remove(node)

            print("Compositor node setup cleared in Stored Scene.")
        else:
            print("Compositor node setup not found in Stored Scene.")


def menu_func(self, context):
    self.layout.operator(CORE_OT_AddCameras.bl_idname)
    self.layout.operator(CORE_OT_DeleteCameras.bl_idname)

def update_handler(scene):
    selected_object = bpy.context.active_object
    if selected_object and selected_object.type == 'CAMERA':
        bpy.context.scene.camera = selected_object
    else:
        bpy.context.scene.camera = None


class CORE_OT_AddFourCameras(Operator):
    bl_label = "Add Four Cameras"
    bl_idname = "object.add_four_cameras"
    bl_description = '- Adds 4 cameras facing the selected object.\n- Select an object first to enable the button'
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        cameras_exist = all(name not in bpy.data.objects for name in ["FrontCamera", "TopCamera", "RightCamera", "LeftCamera"])

        active_object_selected = context.object is not None

        return cameras_exist and active_object_selected

    def execute(self, context):
        global stored_main_scene_name
        active_scene = context.scene
        stored_main_scene_name = active_scene.name + "_Main"  
        
        if context.object is not None:
            selected_object = context.object

            active_scene.name = stored_main_scene_name

            camviews_collection = bpy.data.collections.get("CameraViews")
            if camviews_collection is None:
                camviews_collection = bpy.data.collections.new("CameraViews")
                bpy.context.scene.collection.children.link(camviews_collection)

            cameras = []
            empty = bpy.data.objects.new("CameraEmpty", None)
            camviews_collection.objects.link(empty)
            context.view_layer.objects.active = empty
            empty.location = selected_object.location

            cameras.append(self.create_camera("FrontCamera", (0, -5, 0), (0, 0, 0), selected_object, empty, camviews_collection))
            cameras.append(self.create_camera("TopCamera", (0, 0, 5), (0, 0, 0), selected_object, empty, camviews_collection))
            cameras.append(self.create_camera("RightCamera", (5, 0, 0), (0, 0, 0), selected_object, empty, camviews_collection))
            cameras.append(self.create_camera("LeftCamera", (-5, 0, 0), (0, 0, 0), selected_object, empty, camviews_collection))
            
            for cam in cameras:
                for coll in cam.users_collection:
                    if coll != camviews_collection:
                        coll.objects.unlink(cam)

            # Duplicate the current active scene for each camera
            for i, camera_name in enumerate(["FrontCamera", "TopCamera", "RightCamera", "LeftCamera"]):
                new_scene = active_scene.copy()
                new_scene_name = camera_name

                # Rename the duplicated scene
                new_scene.name = new_scene_name

                # Link existing camera to the new scene
                for cam in cameras:
                    if cam.name not in camviews_collection.objects:
                        camviews_collection.objects.link(cam)
                        cam.parent = empty

                # Set active camera in scene
                new_scene.camera = bpy.data.objects[new_scene_name] 

                # Set camera as active in 3D view
                context.view_layer.objects.active = bpy.data.objects[new_scene_name] 

                bpy.data.objects[new_scene_name].select_set(True)
                bpy.ops.object.select_all(action='DESELECT')

                if camera_name in ["FrontCamera", "TopCamera", "RightCamera", "LeftCamera"]:
                    self.clear_compositor_nodes(bpy.data.scenes[new_scene_name])
                    
                if camera_name == "LeftCamera":
                    # Switch to 'LeftCamera' scene
                    bpy.context.window.scene = bpy.data.scenes[new_scene_name]
                    
                    self.create_compositor(context, new_scene)
                    
            context.window.scene = bpy.data.scenes[stored_main_scene_name]

        self.report({"INFO"}, "Camera Views are added")

        return {'FINISHED'}
    
    def create_camera(self, name, location_offset, rotation, target_object, empty, collection):
        bpy.context.scene.cursor.location = empty.location
        bpy.ops.object.camera_add(enter_editmode=False, align='VIEW', location=(0, 0, 0), rotation=rotation)
        new_camera = bpy.context.object
        new_camera.name = name
        new_camera.location = location_offset

        track_to_constraint = new_camera.constraints.new(type='TRACK_TO')
        track_to_constraint.target = target_object
        track_to_constraint.track_axis = 'TRACK_NEGATIVE_Z'
        track_to_constraint.up_axis = 'UP_Y'

        new_camera.parent = empty

        collection.objects.link(new_camera)

        return new_camera
    
    def clear_compositor_nodes(self, scene):
        if scene.use_nodes:
            tree = scene.node_tree
            for node in tree.nodes:
                tree.nodes.remove(node)
                
    def create_compositor(self, context, scene):
        if context.window.scene == scene:
            scene.use_nodes = True
            tree = scene.node_tree
            for node in tree.nodes:
                tree.nodes.remove(node)
            
            left_camera_scene = bpy.data.scenes.get('LeftCamera')
            if not left_camera_scene:
                # Create 'LeftCamera' scene
                left_camera_scene = bpy.data.scenes.new('LeftCamera')
                context.window.scene = left_camera_scene
                left_camera_scene.use_nodes = True

                # Link objects to the 'LeftCamera' scene
                for obj in context.scene.objects:
                    left_camera_scene.collection.objects.link(obj)

            # Switch to 'LeftCamera' scene before setting up compositor nodes
            if scene.name == "LeftCamera":
                print("LeftCamera scene found")
                context.window.scene = left_camera_scene
                
            # Create compositor nodes for 'LeftCamera'
            self.create_compositor_nodes(tree, "LeftCamera")
            
    def create_compositor_nodes(self, tree, camera_name):
      # FOR FrontCamera
        render_layers_node = tree.nodes.new(type="CompositorNodeRLayers")
        render_layers_node.location = (0, 0)
        render_layers_node.scene = bpy.data.scenes.get("FrontCamera")

        scale_node = tree.nodes.new(type="CompositorNodeScale")
        scale_node.location = (300, 0)
        scale_node.inputs[1].default_value = 0.5
        scale_node.inputs[2].default_value = 0.5

        translate_node = tree.nodes.new(type="CompositorNodeTranslate")
        translate_node.location = (500, 0)
        translate_node.inputs[1].default_value = -480
        translate_node.inputs[2].default_value = 270

        AlphaOver_node = tree.nodes.new(type="CompositorNodeAlphaOver")
        AlphaOver_node.location = (700, -150)
        AlphaOver_node.inputs[0].default_value = 1

        tree.links.new(render_layers_node.outputs["Image"], scale_node.inputs[0])
        tree.links.new(scale_node.outputs["Image"], translate_node.inputs[0])
        tree.links.new(translate_node.outputs["Image"], AlphaOver_node.inputs[1])

      # FOR TopCamera
        render_layers_001_node = tree.nodes.new(type="CompositorNodeRLayers")
        render_layers_001_node.location = (0, -300)
        render_layers_001_node.scene = bpy.data.scenes.get("TopCamera")

        scale_001_node = tree.nodes.new(type="CompositorNodeScale")
        scale_001_node.location = (300, -300)
        scale_001_node.inputs[1].default_value = 0.5
        scale_001_node.inputs[2].default_value = 0.5

        translate_001_node = tree.nodes.new(type="CompositorNodeTranslate")
        translate_001_node.location = (500, -300)
        translate_001_node.inputs[1].default_value = 480
        translate_001_node.inputs[2].default_value = 270

        tree.links.new(render_layers_001_node.outputs["Image"], scale_001_node.inputs[0])
        tree.links.new(scale_001_node.outputs["Image"], translate_001_node.inputs[0])
        tree.links.new(translate_001_node.outputs["Image"], AlphaOver_node.inputs[2])

      # FOR LeftCamera
        render_layers_002_node = tree.nodes.new(type="CompositorNodeRLayers")
        render_layers_002_node.location = (0, -600)
        render_layers_002_node.scene = bpy.data.scenes.get("LeftCamera")

        scale_002_node = tree.nodes.new(type="CompositorNodeScale")
        scale_002_node.location = (300, -600)
        scale_002_node.inputs[1].default_value = 0.5
        scale_002_node.inputs[2].default_value = 0.5

        translate_002_node = tree.nodes.new(type="CompositorNodeTranslate")
        translate_002_node.location = (500, -600)
        translate_002_node.inputs[1].default_value = -480
        translate_002_node.inputs[2].default_value = -270

        AlphaOver_001_node = tree.nodes.new(type="CompositorNodeAlphaOver")
        AlphaOver_001_node.location = (700, -750)
        AlphaOver_001_node.inputs[0].default_value = 1

        tree.links.new(render_layers_002_node.outputs["Image"], scale_002_node.inputs[0])
        tree.links.new(scale_002_node.outputs["Image"], translate_002_node.inputs[0])
        tree.links.new(translate_002_node.outputs["Image"], AlphaOver_001_node.inputs[1])

      # FOR RightCamera
        render_layers_003_node = tree.nodes.new(type="CompositorNodeRLayers")
        render_layers_003_node.location = (0, -900)
        render_layers_003_node.scene = bpy.data.scenes.get("RightCamera")

        scale_003_node = tree.nodes.new(type="CompositorNodeScale")
        scale_003_node.location = (300, -900)
        scale_003_node.inputs[1].default_value = 0.5
        scale_003_node.inputs[2].default_value = 0.5

        translate_003_node = tree.nodes.new(type="CompositorNodeTranslate")
        translate_003_node.location = (500, -900)
        translate_003_node.inputs[1].default_value = 480
        translate_003_node.inputs[2].default_value = -270

        tree.links.new(render_layers_003_node.outputs["Image"], scale_003_node.inputs[0])
        tree.links.new(scale_003_node.outputs["Image"], translate_003_node.inputs[0])
        tree.links.new(translate_003_node.outputs["Image"], AlphaOver_001_node.inputs[2])

        AlphaOver_002_node = tree.nodes.new(type="CompositorNodeAlphaOver")
        AlphaOver_002_node.location = (1000, -400)
        AlphaOver_002_node.inputs[0].default_value = 1

        tree.links.new(AlphaOver_node.outputs["Image"], AlphaOver_002_node.inputs[1])
        tree.links.new(AlphaOver_001_node.outputs["Image"], AlphaOver_002_node.inputs[2])

        # Add OutputFile node
        OutputFile_node = tree.nodes.new(type="CompositorNodeOutputFile")
        OutputFile_node.location = (1200, -600)
        OutputFile_node.label = 'InitialOutputViews'
        if OutputFile_node:
        # Names for the new inputs
            input_names = ["Front_View_", "Top_View_", "Left_View_", "Right_View_"]
        # Add four new inputs to the File Output node with specific names
            for input_name in input_names:
                OutputFile_node.file_slots.new(input_name)
        
        # Link nodes 2 last nodes
        tree.links.new(AlphaOver_002_node.outputs["Image"], OutputFile_node.inputs[0])

        # Add Viewer node
        Viewer_node = tree.nodes.new(type="CompositorNodeViewer")
        Viewer_node.location = (1200, -300)

        # Add Composit node
        Composit_node = tree.nodes.new(type="CompositorNodeComposite")
        Composit_node.location = (1200, -450)
        
        # Link nodes
        tree.links.new(AlphaOver_002_node.outputs["Image"], Viewer_node.inputs[0])
        tree.links.new(AlphaOver_002_node.outputs["Image"], Composit_node.inputs[0])

        tree.links.new(render_layers_node.outputs["Image"], OutputFile_node.inputs[1])
        tree.links.new(render_layers_001_node.outputs["Image"], OutputFile_node.inputs[2])
        tree.links.new(render_layers_002_node.outputs["Image"], OutputFile_node.inputs[3])
        tree.links.new(render_layers_003_node.outputs["Image"], OutputFile_node.inputs[4])

        # Access shared resolution values
        shared_ftlr_resolution_x = bpy.context.window_manager.shared_ftlr_resolution_x
        shared_ftlr_resolution_y = bpy.context.window_manager.shared_ftlr_resolution_y

        return {'FINISHED'}

class CORE_OT_SwitchToScene(Operator):
    bl_label = "To Camera"
    bl_idname = "scene.switch_to"
    
    scene_name: bpy.props.StringProperty()

    def execute(self, context):
        bpy.context.window.scene = bpy.data.scenes[self.scene_name]

        camera_name = self.scene_name
        scene = bpy.context.scene

        # Set hide_viewport only for specific cameras in the current scene
        for obj in scene.objects:
            if obj.type == 'CAMERA' and obj.name in {'FrontCamera', 'TopCamera', 'LeftCamera', 'RightCamera'}:
                obj.hide_viewport = (obj.name != camera_name)

        # Hide cameras with specific names in other scenes
        for obj in bpy.data.objects:
            if obj.type == 'CAMERA' and obj.name in {'FrontCamera', 'TopCamera', 'LeftCamera', 'RightCamera'} and obj.name != camera_name:
                obj.hide_viewport = True

        if camera_name in scene.objects:
            camera_object = scene.objects[camera_name]
            bpy.context.view_layer.objects.active = camera_object
            camera_object.select_set(True)
            bpy.context.object.show_name = True

        # Set the label dynamically based on camera name
        self.bl_label = f"To Camera {camera_name[-1]}"

        return {'FINISHED'}


class CORE_OT_SwitchToFrontCamera(CORE_OT_SwitchToScene):
    bl_idname = "scene.switch_to_front_camera"
    bl_label = "To Front Camera"
    bl_description = '- Switches to the Front Camera where the Depth Of Field and Show Limits can be enabled'
    scene_name: bpy.props.StringProperty(default='FrontCamera')

class CORE_OT_SwitchToTopCamera(CORE_OT_SwitchToScene):
    bl_idname = "scene.switch_to_top_camera"
    bl_label = "To Top Camera"
    bl_description = '- Switches to the Top Camera where the Depth Of Field and Show Limits can be enabled'
    scene_name: bpy.props.StringProperty(default='TopCamera')

class CORE_OT_SwitchToLeftCamera(CORE_OT_SwitchToScene):
    bl_idname = "scene.switch_to_left_camera"
    bl_label = "To Left Camera"
    bl_description = '- Switches to the Left Camera where the Depth Of Field and Show Limits can be enabled'
    scene_name: bpy.props.StringProperty(default='LeftCamera')

class CORE_OT_SwitchToRightCamera(CORE_OT_SwitchToScene):
    bl_idname = "scene.switch_to_right_camera"
    bl_label = "To Right Camera"
    bl_description = '- Switches to the Right Camera where the Depth Of Field and Show Limits can be enabled'
    scene_name: bpy.props.StringProperty(default='RightCamera')


class CORE_OT_SwitchToMainScene(Operator):
    bl_idname = "object.switch_to_main_scene"
    bl_label = "Switch to MAIN SCENE"
    bl_description = '- Switches back to the scene where the object was selected\n- In this scene, the buttons for the Output settings, Render, etc...will be enabled '

    def execute(self, context):
        main_scene_name = None

        for scene in bpy.data.scenes:
            if '_Main' in scene.name:
                main_scene_name = scene.name
                break

        if main_scene_name:
            # Switch to the main scene
            context.window.scene = bpy.data.scenes[main_scene_name]

            # Set hide_viewport only for specific cameras in the current scene
            for obj in context.scene.objects:
                if obj.type == 'CAMERA' and obj.name in {'FrontCamera', 'TopCamera', 'LeftCamera', 'RightCamera'}:
                    obj.hide_viewport = False

            # Hide cameras with specific names in other scenes
            for obj in bpy.data.objects:
                if obj.type == 'CAMERA' and obj.name in {'FrontCamera', 'TopCamera', 'LeftCamera', 'RightCamera'}:
                    obj.hide_viewport = False

            self.report({'INFO'}, f"Switched to main scene: {main_scene_name}")
            return {'FINISHED'}

        self.report({'ERROR'}, "No scene with '_Main' in the name found.")
        return {'CANCELLED'}


class CORE_OT_ftlr_SetInstaResolution(Operator):
    bl_idname = "render.set_ftlr_insta_resolution"
    bl_label = "Toggle Resolution"
    bl_description = "- Toggle resolution between 1080x1350 and 1920x1080\n- 1920x1080 is default"
    
    def execute(self, context):
        if bpy.context.window_manager.shared_ftlr_resolution_x == 1080 and bpy.context.window_manager.shared_ftlr_resolution_y == 1350:
            bpy.context.window_manager.shared_ftlr_resolution_x = 1920
            bpy.context.window_manager.shared_ftlr_resolution_y = 1080
        else:
            bpy.context.window_manager.shared_ftlr_resolution_x = 1080
            bpy.context.window_manager.shared_ftlr_resolution_y = 1350
        return {'FINISHED'}

class CORE_ftlr_RenderPaths(PropertyGroup):
    render_ftlr_output_directory: bpy.props.StringProperty(
        name="Render Output Directory",
        subtype='DIR_PATH',
        default="//render_output",
    )
    shared_ftlr_resolution_x: bpy.props.IntProperty(
        name="Shared Resolution X",
        default=1920,
        min=1
    )
    shared_ftlr_resolution_y: bpy.props.IntProperty(
        name="Shared Resolution Y",
        default=1080,
        min=1
    )
#########################################
# Operator to change the File Output path
#########################################    
# 
def find_scene_with_ftlr_output_node_label():
    for scene in bpy.data.scenes:
        if scene.use_nodes:
            for node in scene.node_tree.nodes:
                if node.label == 'InitialOutputViews' or node.label == 'NewOutputViews':
                    return scene
    return None
    
class CORE_OT_ftlr_CustomRenderOutput(Operator):
    bl_idname = "render.custom_ftlr_render_output"
    bl_label = "Set Custom Render Output"
    bl_description = '- Confirm the image name, render output path and resolution '
    
    @classmethod
    def poll(cls, context):
        camera_names = ["FrontCamera", "TopCamera", "RightCamera", "LeftCamera"]

        for scene_name in camera_names + [stored_main_scene_name]:
            if scene_name in bpy.data.scenes:
                for camera_name in camera_names:
                    if camera_name in bpy.data.objects:
                        return True 
        return False  

    def execute(self, context):

        # Set the resolution for scenes 
        shared_ftlr_resolution_x = bpy.context.window_manager.shared_ftlr_resolution_x
        shared_ftlr_resolution_y = bpy.context.window_manager.shared_ftlr_resolution_y
      
        for new_scene_name in ["FrontCamera", "TopCamera", "RightCamera", "LeftCamera"]:
            scene = bpy.data.scenes.get(new_scene_name)
            if scene:
                scene.render.resolution_x = shared_ftlr_resolution_x
                scene.render.resolution_y = shared_ftlr_resolution_y  
                scene.render.resolution_percentage = 100     
                    
        bpy.data.scenes["LeftCamera"].render.resolution_percentage        
        # Set hide_viewport only for specific cameras in the current scene
        for obj in context.scene.objects:
            if obj.type == 'CAMERA' and obj.name in {'FrontCamera', 'TopCamera', 'LeftCamera', 'RightCamera'}:
                obj.hide_viewport = False

        # Hide cameras with specific names in other scenes
        for obj in bpy.data.objects:
            if obj.type == 'CAMERA' and obj.name in {'FrontCamera', 'TopCamera', 'LeftCamera', 'RightCamera'}:
                obj.hide_viewport = False
                      
        scene = bpy.context.scene
        
        # Set 'LeftCamera' scene as the active scene
        left_camera_scene = bpy.data.scenes.get('LeftCamera')
        if left_camera_scene:
            context.window.scene = left_camera_scene
        
        scene.render.filepath = bpy.context.window_manager.shared_ftlr_render_path
       
        bpy.context.scene.use_nodes = True
        tree = bpy.context.scene.node_tree
        tree_type = 'COMPOSITING'
        
        for node in tree.nodes:
            if node.bl_idname == 'CompositorNodeTranslate':
                tree.nodes.remove(node)

        # Find and delete existing 'File Output' nodes
        for node in tree.nodes:
            if node.type == 'OUTPUT_FILE' and (node.label == 'InitialOutputViews' or node.label == 'NewOutputViews'): 
                tree.nodes.remove(node)
                
        # Add a new 'File Output' node
        NewOutputFile_node = tree.nodes.new(type='CompositorNodeOutputFile')
        NewOutputFile_node.label = 'NewOutputViews' 
        NewOutputFile_node.location = (1200, -600)
        NewOutputFile_node.base_path = bpy.context.window_manager.shared_ftlr_render_path
        NewOutputFile_node.file_slots[0].path = bpy.context.window_manager.shared_ftlr_custom_image_name + "_Combined_Image_"
        
        if NewOutputFile_node:
            # Names for the new inputs
                input_names = ["Front_View_", "Top_View_", "Left_View_", "Right_View_"]
            # Add four new inputs to the File Output node with specific names
                for input_name in input_names:
                    NewOutputFile_node.file_slots.new(bpy.context.window_manager.shared_ftlr_custom_image_name + "_" + input_name)

        # Add four 'Translate' nodes
        translate_nodes = []
        translate_values = [
            (-shared_ftlr_resolution_x / 4, shared_ftlr_resolution_y / 4, 0),
            (shared_ftlr_resolution_x / 4, shared_ftlr_resolution_y / 4, 0),
            (-shared_ftlr_resolution_x / 4, -shared_ftlr_resolution_y / 4, 0),
            (shared_ftlr_resolution_x / 4, -shared_ftlr_resolution_y / 4, 0),
        ]

        for i in range(4):
            translate_node = tree.nodes.new(type='CompositorNodeTranslate')
            translate_node.location = (500, -i * 200)
            translate_node.inputs[1].default_value = translate_values[i][0]  # X translation
            translate_node.inputs[2].default_value = translate_values[i][1]  # Y translation
            translate_nodes.append(translate_node)  

        # Link nodes
        link_ftlr_nodes(context)

        self.update_translate_nodes(tree, shared_ftlr_resolution_x, shared_ftlr_resolution_y)
        
        scene = bpy.context.scene
        # Set 'LeftCamera' scene as the active scene
        left_camera_scene = bpy.data.scenes.get('LeftCamera')
        if left_camera_scene:
            context.window.scene = left_camera_scene

        self.report({"INFO"}, "Output settings adjusted")

        return {'FINISHED'}

    def update_translate_nodes(self, tree, shared_ftlr_resolution_x, shared_ftlr_resolution_y):
        translate_nodes = [node for node in tree.nodes if node.type == 'CompositorNodeTranslate']

        for i, translate_node in enumerate(translate_nodes):
            translate_node.inputs[1].default_value = -shared_ftlr_resolution_x / 4 if i % 2 == 0 else shared_ftlr_resolution_x / 4
            translate_node.inputs[2].default_value = shared_ftlr_resolution_y / 4 if i // 2 == 0 else -shared_ftlr_resolution_y / 4
                 
        return {'FINISHED'}   
    
     
class CORE_OT_TrackSelectedObject(Operator):
    bl_label = "Track Selected Object"
    bl_idname = "object.track_selected_object"
    bl_description = "The cameras will track the newly selected object when pressed"
    bl_options = {'REGISTER', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        main_scene_name = None

        for scene in bpy.data.scenes:
            if '_Main' in scene.name:
                main_scene_name = scene.name
                break

        if bpy.context.scene.name != main_scene_name:
            return False

        camera_names = ["FrontCamera", "TopCamera", "RightCamera", "LeftCamera"]
        for scene_name in camera_names + [main_scene_name]:
            scene = bpy.data.scenes.get(scene_name)
            if scene:
                for camera_name in camera_names:
                    if camera_name in bpy.data.objects:
                        return True

        return False

    def execute(self, context):
        if context.object is None:
            self.report({'ERROR'}, "No object selected")
            return {'CANCELLED'}

        selected_object = context.object

        camera_scenes = ["FrontCamera", "TopCamera", "RightCamera", "LeftCamera"]
        for scene_name in camera_scenes:
            scene = bpy.data.scenes.get(scene_name)
            if scene:
                # Get camera and set track-to constraint to selected object
                camera = bpy.data.objects.get(scene_name)
                if camera:
                    track_to_constraint = camera.constraints.get("Track To")
                    if track_to_constraint:
                        track_to_constraint.target = selected_object

        return {'FINISHED'}

class CORE_OT_MoveCursorToOrigin(Operator):
    bl_label = "Move Cursor to Origin and Set Empty Position"
    bl_idname = "object.move_cursor_to_origin"
    bl_description = '- Pressing this button will position the cameras to the selected object '
    bl_options = {'REGISTER', 'UNDO'}
       
    @classmethod
    def poll(cls, context):
        main_scene_name = None
        for scene in bpy.data.scenes:
            if '_Main' in scene.name:
                main_scene_name = scene.name
                break

        if bpy.context.scene.name != main_scene_name:
            return False
        
        camera_names = ["FrontCamera", "TopCamera", "RightCamera", "LeftCamera"]
        for scene_name in camera_names + [stored_main_scene_name]:
            if scene_name in bpy.data.scenes:
                for camera_name in camera_names:
                    if camera_name in bpy.data.objects:
                        return True
        return False
   
    def execute(self, context):
        if context.object is not None:
            selected_object = context.object

            bpy.context.scene.cursor.location = selected_object.location

            empty = bpy.data.objects.get("CameraEmpty")
            if empty:
                empty.location = bpy.context.scene.cursor.location

        return {'FINISHED'}
           
##########################################################################
# Operator that renders the image of all 4 cameras
########################################################################## 
class CORE_OT_ftlr_RenderButton(bpy.types.Operator):
    bl_idname = "render.render_ftlr_button"
    bl_label = "RENDER"
    bl_description = "Renders the 4 camera views and a combined image of the views"

    @classmethod
    def poll(cls, context):
        left_camera_scene = bpy.data.scenes.get('LeftCamera')
        if left_camera_scene:
            compositor = left_camera_scene.node_tree
            for node in compositor.nodes:
                if node.type == 'OUTPUT_FILE' and node.label == 'NewOutputViews':
                    return True       
        return False  
             
    def execute(self, context):
        # Set the scene to 'LeftCamera' scene
        bpy.context.window.scene = bpy.data.scenes['LeftCamera']                     
        bpy.ops.render.render("INVOKE_DEFAULT")  

        # Register render complete callback function
        bpy.app.handlers.render_complete.append(self.open_ftlr_render_output_directory)
        
        return {'RUNNING_MODAL'} 

    def open_ftlr_render_output_directory(self, *args):
        # This function is called when rendering is completed
        print("Rendering completed. Opening render output directory...")
        render_ftlr_output_directory = bpy.path.abspath(bpy.context.window_manager.shared_ftlr_render_path)

        try:
            if bpy.context.window_manager.ftlr_open_folder_after_render:
                if os.name == 'nt':  # Windows
                    subprocess.Popen(['explorer', render_ftlr_output_directory], shell=True)
                elif os.name == 'posix':  # Linux and macOS
                    subprocess.Popen(['xdg-open', render_ftlr_output_directory])
                else:
                    print("Unsupported operating system.")
        except Exception as e:
            print(f"Error opening file explorer: {e}")

        # Unregister the callback function
        bpy.app.handlers.render_complete.remove(self.open_ftlr_render_output_directory)
        
        return {'FINISHED'}   
     
    def bl_ftlr_update(self, context):
        bpy.app.handlers.render_complete.append(self.open_ftlr_render_output_directory) 

class CORE_OT_ftlr_DeleteCameras(bpy.types.Operator):
    bl_label = "Delete Cameras"
    bl_idname = "object.delete_ftlr_cameras"
    bl_description = '- Deletes the 4 added cameras, linked empty, and CameraViews collections in specified scenes'
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        camera_names = ["FrontCamera", "TopCamera", "RightCamera", "LeftCamera"]

        for scene_name in camera_names + [stored_main_scene_name]:
            if scene_name in bpy.data.scenes:
                for camera_name in camera_names:
                    if camera_name in bpy.data.objects:
                        return True  
        return False  
    
    def remove_main_suffix(self, name):
        if name.endswith('_Main'):
            return name[:-5]
        return name

   
    def execute(self, context):
        camera_names = ["FrontCamera", "TopCamera", "RightCamera", "LeftCamera"]

        for scene in bpy.data.scenes:
            scene.name = self.remove_main_suffix(scene.name)
            
            # Delete cameras
            cameras = [obj for obj in scene.objects if obj.name in camera_names]
            bpy.data.batch_remove(cameras)

            # Delete linked empty
            linked_empty = scene.objects.get("CameraEmpty")
            if linked_empty:
                bpy.data.batch_remove([linked_empty])

            # Delete CamViews collection
            camviews_collection = scene.collection.children.get("CameraViews")
            if camviews_collection:
                bpy.data.batch_remove([camviews_collection])

            # Delete collections starting with 'CameraViews'
            for collection in scene.collection.children:
                if collection.name.startswith("CameraViews"):
                    bpy.data.batch_remove([collection])
                    
            # Delete scenes with specified names
            if scene.name in camera_names:
                bpy.data.scenes.remove(scene)
                
        # Purge orphan data
        bpy.ops.outliner.orphans_purge(do_local_ids=True, do_linked_ids=True, do_recursive=True)

        return {'FINISHED'}
 
################################################################    
################### Add-ons Preferences Update Panel ###########

panels = (
               
        CORE_PT_cameras,
        CORE_PT_ftlr_cameras,
        
        )
        
def update_panel(self, context):
    message = ": Updating Panel locations has failed"
    try:
        for panel in panels:
            if "bl_rna" in panel.__dict__:
                bpy.utils.unregister_class(panel)

        for panel in panels:
            panel.bl_category = context.preferences.addons[__name__].preferences.category
            bpy.utils.register_class(panel)

    except Exception as e:
        print("\n[{}]\n{}\n\nError:\n{}".format(__name__, message, e))
        pass        
        
class ADDON_PREFERENCES(AddonPreferences):
    # This must match the addon name, use '__package__'
    # when defining this in a submodule of a python package.
    bl_idname = __name__

    category: StringProperty(
            name="Tab Category",
            description="Choose a name for the category of the panel",
            default="ComRen",
            update=update_panel
            )

    def draw(self, context):
        layout = self.layout
        row = layout.row()
        col = row.column()
        col.label(text="Tab Category:")
        col.prop(self, "category", text="")
        
##########################################################################################################    
# REGISTER
##########################################################################################################
##########################################################################################################
## REGISTER
##########################################################################################################
classes = [
    ADDON_PREFERENCES,
    CORE_PT_cameras,
    CORE_OT_SwitchScene,
    CORE_OT_MultiCopyScene,
    CORE_OT_AddCameras,
    CORE_RenderPaths,
    CORE_OT_SetInstaResolution,
    CORE_OT_CustomRenderOutput,
    CORE_OT_RenderButton,
    CORE_OT_DeleteCameras,
    CORE_ftlr_RenderPaths,
    CORE_PT_ftlr_cameras,
    CORE_OT_AddFourCameras,
    CORE_OT_SwitchToScene,
    CORE_OT_SwitchToFrontCamera,
    CORE_OT_SwitchToTopCamera,
    CORE_OT_SwitchToLeftCamera,
    CORE_OT_SwitchToRightCamera,
    CORE_OT_SwitchToMainScene,
    CORE_OT_TrackSelectedObject,
    CORE_OT_MoveCursorToOrigin,
    CORE_OT_ftlr_SetInstaResolution,
    CORE_OT_ftlr_CustomRenderOutput,
    CORE_OT_ftlr_RenderButton,
    CORE_OT_ftlr_DeleteCameras

]

###############   REGISTER ICONS
def register_icons():       
    icons = ["3COPY", "ADDCAM", "DELCAM", "SETTINGS", "RESOL", "FOURSCENES", "FTLR"] 
    bpy.types.Scene.Combined_Render_icons = bpy.utils.previews.new()
# Path to the folder where the icon is
# the path is calculated relative to this .py file inside the addon folder
    icons_dir = os.path.join( os.path.dirname( __file__ ), "icons" )
# Load a preview thumbnail of a file and store in the previews collection
    for icon in icons:
        bpy.types.Scene.Combined_Render_icons.load( icon, os.path.join( icons_dir, icon + ".png" ), 'IMAGE' )
                
###############   UNREGISTER ICONS
def unregister_icons():
    bpy.utils.previews.remove( bpy.types.Scene.Combined_Render_icons )

###############   REGISTER ADDON
def register():
    register_icons() 
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)
    bpy.types.VIEW3D_MT_mesh_add.append(menu_func)
    bpy.types.Scene.cam_empty_distance = StringProperty(
        name="",
        default="3000",
        description="- Distance between camera and empty, in millimeters"
    )
    bpy.types.Scene.use_for_cameras = BoolProperty(
        name="Use for Cameras",
        default=True,
        description="- Include this scene for adding cameras"
    )
    bpy.types.Scene.custom_image_name = StringProperty(
        name="Custom Image Name",
        default="Image",
        update=update_filename
    )
    bpy.types.Scene.custom_ftlr_image_name = StringProperty(
        name="Custom Image Name",
        default="Image",
        update=update_ftlr_filename
    ) 
    bpy.types.WindowManager.open_folder_after_render = BoolProperty(
        name="Open Folder After Render",
        description="Open the render folder after rendering",
        default=True
    )
    bpy.types.WindowManager.ftlr_open_folder_after_render = BoolProperty(
        name="Open Folder After Render",
        description="Open the render folder after rendering",
        default=True
    )
    bpy.types.Scene.cameras_added = BoolProperty(default=False)
    bpy.types.Scene.paths = PointerProperty(type=CORE_RenderPaths)
    bpy.types.Scene.paths = PointerProperty(type=CORE_ftlr_RenderPaths)
    bpy.app.handlers.depsgraph_update_post.append(update_handler)
    update_panel(None, bpy.context)
    bpy.types.WindowManager.active_scene_name = bpy.props.StringProperty() 
    bpy.app.handlers.load_post.append(CORE_OT_RenderButton.bl_rna_update)
    bpy.app.handlers.load_post.append(CORE_OT_ftlr_RenderButton.bl_ftlr_update)

###############   UNREGISTER ADDON
def unregister():
    unregister_icons()
    from bpy.utils import unregister_class
    for cls in reversed(classes):
        unregister_class(cls)  
    bpy.types.VIEW3D_MT_mesh_add.remove(menu_func)
    del bpy.types.Scene.cam_empty_distance
    del bpy.types.Scene.use_for_cameras
    del bpy.types.Scene.paths
    del bpy.types.Scene.cameras_added 
    del bpy.types.WindowManager.open_folder_after_render
    del bpy.types.WindowManager.ftlr_open_folder_after_render
    del bpy.types.WindowManager.active_scene_name
    if update_handler in bpy.app.handlers.depsgraph_update_post:
        bpy.app.handlers.depsgraph_update_post.remove(update_handler)
    
    
if __name__ == "__main__":
    register()

#################################################################

